#!/bin/python

license = """
/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 
 NOTE: The dsm-api library and it's script bindings are unsupported interfaces and
 subject to change without further notice.
 
 */
 """

import sys
from xml.etree.ElementTree import ElementTree
from optparse import OptionParser
from string import Template

from Model import *
from CGenerator import CGenerator
from PyGenerator import PyGenerator
from LuaGenerator import LuaGenerator
from ApiLoggerGenerator import ApiLoggerGenerator

def parseCalls(xml_part, calls, message_ids, prefixes):
    for api_call in xml_part.findall("dsm-call"):
        message_id = MessageId(api_call.attrib["name"], 
                               api_call.attrib["message_id"])
        message_ids.append(message_id)
        messageIdDeprecated = api_call.attrib["deprecated"] == "True" if api_call.attrib.has_key("deprecated") else False
        
        if message_id.get_prefix() in prefixes:
            sys.stderr.write(message_id.get_name() + " leads to a duplicate prefix with " + prefixes[message_id.get_prefix()] + "\n")
            exit()
        else:
            prefixes[message_id.get_prefix()]= message_id.get_name()
        
        timeouts = api_call.findall("timeout")
        if (timeouts and len(timeouts) > 0):
            message_id.set_timeout(int(timeouts[0].attrib["seconds"]))
        else:
            message_id.set_timeout(0)
        
        modifiers = api_call.findall("modifier")
        common_request = api_call.findall("request/parameter")
        common_response = api_call.findall("response/parameter")
        common_description = api_call.findall("description")
        common_error = api_call.findall("error")
        common_description = (str(common_description[0].text) + " ") if (len(common_description) > 0 and common_description[0].text) else ""
        
        if (len(modifiers) == 0):
            calls.append(ApiCall(message_id.get_name(), 
                                 message_id.get_id(), 
                                 map(Parameter, common_request), 
                                 map(Parameter, common_response), 
                                 message_id.get_timeout(),
                                 0, 
                                 common_description,
                                 messageIdDeprecated))
        else:
            for modifier in modifiers:
                modifier_obj = message_id.add_modifier(modifier)
                modifierIdDeprecated = messageIdDeprecated or (modifier.attrib["deprecated"] == "true" if modifier.attrib.has_key("deprecated") else False)
                description = modifier.findall("description")
                documentation = common_description
                
                mod_timeouts = modifier.findall("timeout")
                if (mod_timeouts and len(mod_timeouts) > 0):
                    mod_timeout = int(mod_timeouts[0].attrib["seconds"])
                else:
                    mod_timeout = message_id.get_timeout();
                                  
                documentation += str(description[0].text) if (len(description) > 0 and description[0].text) else ""
                
                calls.append(ApiCall(message_id.get_name() + "_" + modifier.attrib["identifier"], 
                                     message_id.get_id(),
                                     map(Parameter, common_request + modifier.findall("request/parameter")),
                                     map(Parameter, common_response + modifier.findall("response/parameter")),
                                     mod_timeout,
                                     modifier_obj.get_id(),
                                     documentation,
                                     modifierIdDeprecated))
    

def parseEvents(xml_part, events, message_ids, prefixes):
    for api_event in xml_part.findall("dsm-event"):
        message_id = MessageId(api_event.attrib["name"], 
                               api_event.attrib["message_id"])
        message_ids.append(message_id)
        messageIdDeprecated = api_event.attrib["deprecated"] == "True" if api_event.attrib.has_key("deprecated") else False
        
        if message_id.get_prefix() in prefixes:
            sys.stderr.write(message_id.get_name() + " leads to a duplicate prefix with " + prefixes[message_id.get_prefix()] + "\n")
            exit()
        else:
            prefixes[message_id.get_prefix()]= message_id.get_name()
            
        common_response = api_event.findall("response/parameter")
        common_description = api_event.findall("description")
        modifiers = api_event.findall("modifier")
        common_description = (str(common_description[0].text) + " ") if (len(common_description) > 0 and common_description[0].text) else ""
        
        if (len(modifiers) == 0):
            events.append(ApiCall(message_id.get_name(), 
                                  message_id.get_id(), 
                                  None, 
                                  map(Parameter, common_response), 
                                  0,
                                  0, 
                                  common_description,
                                  messageIdDeprecated))
        else:
            for modifier in modifiers:
                modifier_obj = message_id.add_modifier(modifier)
                modifierIdDeprecated = messageIdDeprecated or (modifier.attrib["deprecated"] == "True" if modifier.attrib.has_key("deprecated") else False)
                description = modifier.findall("description")
                documentation = common_description
                                
                documentation += str(description[0].text) if (len(description) > 0 and description[0].text) else ""
                
                events.append(ApiCall(message_id.get_name() + "_" + modifier.attrib["identifier"], 
                                      message_id.get_id(),
                                      None,
                                      map(Parameter, common_response + modifier.findall("response/parameter")),
                                      0,
                                      modifier_obj.get_id(),
                                      documentation,
                                      modifierIdDeprecated))
                  
header_clause = license + """/*
 * DO NOT MANUALLY MODIFY! 
 *
 * THIS FILE IS AUTOGENERATED FROM THE dSM-API XML DEFINITIONS IN
 * dsm-api.xml.
 * Make any changes in the XML definition and regenerate this file.
 */
"""

if __name__ == "__main__":
    parser = OptionParser()
    parser.add_option("-i", "--input", dest="infile",
                      help="XML input")
    parser.add_option("-s", "--schema", dest="schema",
                      help="XML schema definition")
    parser.add_option("-d", "--definition", dest="header",
                      help="C header output file")
    parser.add_option("-c", "--code", dest="code",
                      help="C source code output file")
    parser.add_option("-e", "--enums", dest="enums",
                      help="C enum header output file")
    parser.add_option("-p", "--swig-python", dest="swig_python",
                      help="SWIG output file for Python bindings")
    parser.add_option("-l", "--swig-lua", dest="swig_lua",
                      help="SWIG output file for Lua bindings")
    parser.add_option("-L", "--api-logger", dest="logger",
                      help="dsm-api logger (on python basis)")
    parser.add_option("-v", "--version", dest="version",
                      help="API version")
    
    (options, args) = parser.parse_args()
    
    try:
        from lxml import etree
        try:
            xmlschema_doc = etree.parse(options.schema)
            xmlschema = etree.XMLSchema(xmlschema_doc)
            
            etree_doc = etree.parse(options.infile)
            
            xmlschema.assertValid(etree_doc)
        except etree.DocumentInvalid as e:
            print e
            sys.stderr.write("The XML file: " + options.infile + "is not valid according to the schema definition\n")
            exit(-1)
    except:
        print "lxml not installed. The XML file", options.infile, "is not validated against the schema."
        pass
    
    tree = ElementTree()
    doc = tree.parse(options.infile)
    
    groups = list()
    message_ids = list()
    prefixes = dict()
    error_list = dict()
    length_enum = dict()
    
    errors = map(Error, doc.findall("error"))
    for error in errors:
        error_list[int(error.getValue(), 16)] = error
        
    main_page = doc.findall("description")[0].text
    
    # parse grouped calls
    for group in doc.findall("group"):
        group_obj = CallGroup(group.attrib["name"], group.findall("description")[0].text)
        groups.append(group_obj)
        parseCalls(group, group_obj.calls, message_ids, prefixes)
        parseEvents(group, group_obj.events, message_ids, prefixes)
    
    #collect all ungrouped calls and events
    group_obj = CallGroup("Ungrouped")
    parseCalls(doc, group_obj.calls, message_ids, prefixes)
    parseEvents(doc, group_obj.events, message_ids, prefixes)
    if len(group_obj.calls) != 0 or len(group_obj.events) != 0:
        groups.append(group_obj)
     
    for group in groups:
        for call in group.calls + group.events:
            call.fillLengthEnum(length_enum)

    c_generator = CGenerator(header_clause, groups, options.version)
    c_generator.header(main_page, options.header)
    c_generator.implementation(options.code)
    c_generator.const_header(options.enums, message_ids, error_list, length_enum)
    
    if options.swig_python:
        py_generator = PyGenerator(header_clause, groups)
        py_generator.swig(options.swig_python, length_enum)
        
        if options.logger:
            logger = ApiLoggerGenerator(header_clause, groups)
            logger.logger(options.logger)
    
    if options.swig_lua:
        lua_generator = LuaGenerator(header_clause, groups)
        lua_generator.swig(options.swig_lua, length_enum)
